from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Literal, Optional, Sequence

Granularity = str  # "repo" | "cwd" | "path:N"


@dataclass(frozen=True)
class ScopeInfo:
    cwd: Path
    repo_root: Optional[Path]
    repo_slug: str
    rel_path: str
    scope_type: Literal["global", "project"]
    scope_id: str


REPO_ROOT_MARKERS: Sequence[str] = ("pyproject.toml", "package.json", "go.mod")


def find_repo_root(start: Path) -> Optional[Path]:
    start = start.resolve()
    candidates: Iterable[Path] = (start, *start.parents)

    for candidate in candidates:
        git_marker = candidate / ".git"
        if git_marker.exists():
            return candidate

    for candidate in candidates:
        for marker in REPO_ROOT_MARKERS:
            p = candidate / marker
            if marker.endswith("/") and p.is_dir():
                return candidate
            if p.exists():
                return candidate

    return None


def _safe_rel_path(repo_root: Path, cwd: Path) -> str:
    try:
        rel = cwd.resolve().relative_to(repo_root.resolve())
    except Exception:
        return ""
    if str(rel) == ".":
        return ""
    return rel.as_posix()


def _scope_id_for(rel_path: str, repo_slug: str, granularity: Granularity) -> str:
    if not repo_slug:
        return ""

    if rel_path == "":
        return repo_slug

    if granularity == "repo":
        return repo_slug

    if granularity == "cwd":
        return f"{repo_slug}::{rel_path}"

    if granularity.startswith("path:"):
        n_str = granularity.split(":", 1)[1]
        try:
            n = int(n_str)
        except ValueError:
            n = 0

        if n <= 0:
            return repo_slug

        parts = [p for p in rel_path.split("/") if p]
        prefix = "/".join(parts[:n])
        if prefix == "":
            return repo_slug
        return f"{repo_slug}::{prefix}"

    return repo_slug


def derive_scope(
    cwd: Path,
    *,
    repo_root_override: Optional[Path] = None,
    scope_type_override: Optional[Literal["global", "project"]] = None,
    scope_id_override: Optional[str] = None,
    granularity: Granularity = "repo",
) -> ScopeInfo:
    cwd = cwd.resolve()
    repo_root = repo_root_override.resolve() if repo_root_override else find_repo_root(cwd)

    repo_slug = repo_root.name if repo_root else ""
    rel_path = _safe_rel_path(repo_root, cwd) if repo_root else ""

    inferred_scope_type: Literal["global", "project"] = "project" if repo_root else "global"
    scope_type: Literal["global", "project"] = scope_type_override or inferred_scope_type

    if scope_type == "global":
        scope_id = scope_id_override or "global"
        return ScopeInfo(
            cwd=cwd,
            repo_root=repo_root,
            repo_slug=repo_slug,
            rel_path=rel_path,
            scope_type=scope_type,
            scope_id=scope_id,
        )

    # project scope
    if not repo_root:
        # explicit project scope without a repo root: treat cwd as pseudo-root.
        repo_root = cwd
        repo_slug = cwd.name
        rel_path = ""

    scope_id = scope_id_override or _scope_id_for(rel_path, repo_slug, granularity)

    return ScopeInfo(
        cwd=cwd,
        repo_root=repo_root,
        repo_slug=repo_slug,
        rel_path=rel_path,
        scope_type=scope_type,
        scope_id=scope_id,
    )


def scope_chain(scope: ScopeInfo, *, include_parents: bool = True) -> list[str]:
    if scope.scope_type != "project":
        return [scope.scope_id]

    if not include_parents:
        return [scope.scope_id]

    sid = scope.scope_id
    if "::" not in sid:
        return [sid]

    repo_slug, rest = sid.split("::", 1)
    repo_slug = repo_slug.strip()
    rest = rest.strip().strip("/")

    if not rest:
        return [repo_slug]

    parts = [p for p in rest.split("/") if p]
    chain: list[str] = []
    for i in range(len(parts), 0, -1):
        chain.append(f"{repo_slug}::{'/'.join(parts[:i])}")
    chain.append(repo_slug)
    return chain



from __future__ import annotations

from pathlib import Path

import pytest

from vibemem.config import VibememConfig, load_config
from vibemem.store.base import ConfigError


def test_load_config_raises_on_malformed_json(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    cfg_path = tmp_path / "config.json"
    cfg_path.write_text("{", encoding="utf-8")

    monkeypatch.setattr(VibememConfig, "config_file", staticmethod(lambda: cfg_path))

    with pytest.raises(ConfigError):
        load_config()




from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import weaviate
from weaviate.auth import AuthApiKey
from weaviate.classes.config import Configure, DataType, Property
from weaviate.collections.classes import filters as wfilters
from weaviate.collections.classes import grpc

from vibemem.config import VibememConfig
from vibemem.embeddings import EmbeddingClient, EmbeddingError
from vibemem.models import (
    _now_iso,
    ListFilters,
    Memory,
    MemoryUpdate,
    SearchHit,
    memory_from_properties,
    memory_to_weaviate_properties,
)
from vibemem.store.base import ConnectionError, NotFoundError, SearchRequest


@dataclass(frozen=True)
class _ParsedEndpoint:
    host: str
    port: int
    secure: bool


def _parse_host_port(url: str, *, default_port_http: int, default_port_https: int) -> _ParsedEndpoint:
    from urllib.parse import urlparse

    url = url.strip()
    # Accept host:port inputs by assuming http://
    if "://" not in url:
        url = "http://" + url

    parsed = urlparse(url)
    scheme = parsed.scheme.lower() if parsed.scheme else "http"
    secure = scheme == "https"

    host = parsed.hostname or ""
    if not host:
        raise ValueError(f"Invalid URL (missing host): {url}")

    if parsed.port is not None:
        port = int(parsed.port)
    else:
        port = default_port_https if secure else default_port_http

    return _ParsedEndpoint(host=host, port=port, secure=secure)


def _looks_like_cloud_url(url: str) -> bool:
    try:
        from urllib.parse import urlparse

        url = url.strip()
        if "://" not in url:
            url = "https://" + url
        host = (urlparse(url).hostname or "").lower()
    except Exception:
        return False
    return host.endswith(".weaviate.cloud") or host.endswith(".weaviate.network") or host.endswith(".wcs.api.weaviate.io")


def _missing_vectorizer_error(err: Exception) -> bool:
    msg = str(err)
    return "VectorFromInput" in msg and "without vectorizer" in msg


class WeaviateStore:
    def __init__(self, cfg: VibememConfig) -> None:
        self._cfg = cfg
        self._collection_name = cfg.weaviate_collection
        self._embedder: Optional[EmbeddingClient] = None

        if not cfg.weaviate_url:
            raise ConnectionError("Weaviate URL is not configured (set VIBEMEM_WEAVIATE_URL).")

        if cfg.embedding_host and cfg.embedding_port:
            self._embedder = EmbeddingClient(
                host=cfg.embedding_host,
                port=int(cfg.embedding_port),
                model=cfg.embedding_model,
                timeout_s=cfg.request_timeout_s,
            )

        timeout = weaviate.config.Timeout(
            query=cfg.request_timeout_s,
            insert=max(cfg.request_timeout_s, 20.0),
            init=min(cfg.request_timeout_s, 10.0),
        )
        additional_config = weaviate.config.AdditionalConfig(timeout=timeout, trust_env=True)

        auth_credentials = None
        if cfg.weaviate_api_key:
            auth_credentials = AuthApiKey(api_key=cfg.weaviate_api_key.get_secret_value())

        try:
            if _looks_like_cloud_url(cfg.weaviate_url):
                if not auth_credentials:
                    raise ConnectionError("Weaviate Cloud URL provided but API key is missing (VIBEMEM_WEAVIATE_API_KEY).")
                self._client = weaviate.connect_to_weaviate_cloud(
                    cluster_url=cfg.weaviate_url,
                    auth_credentials=auth_credentials,
                    additional_config=additional_config,
                    skip_init_checks=False,
                )
            else:
                http = _parse_host_port(cfg.weaviate_url, default_port_http=8080, default_port_https=443)
                if cfg.weaviate_grpc_url:
                    grpc_ep = _parse_host_port(cfg.weaviate_grpc_url, default_port_http=50051, default_port_https=443)
                else:
                    grpc_ep = _ParsedEndpoint(host=http.host, port=50051, secure=http.secure)

                self._client = weaviate.connect_to_custom(
                    http_host=http.host,
                    http_port=http.port,
                    http_secure=http.secure,
                    grpc_host=grpc_ep.host,
                    grpc_port=grpc_ep.port,
                    grpc_secure=grpc_ep.secure,
                    auth_credentials=auth_credentials,
                    additional_config=additional_config,
                    skip_init_checks=False,
                )

            self._ensure_collection()
            self._collection = self._client.collections.use(self._collection_name)
        except Exception as e:
            raise ConnectionError(f"Failed to connect to Weaviate: {e}") from e

    def close(self) -> None:
        try:
            self._client.close()
        except Exception:
            return

    def _ensure_collection(self) -> None:
        try:
            exists = self._client.collections.exists(self._collection_name)
        except Exception as e:
            raise ConnectionError(f"Cannot check Weaviate schema: {e}") from e

        if exists:
            return

        properties = [
            Property(name="text", data_type=DataType.TEXT),
            Property(name="mem_type", data_type=DataType.TEXT),
            Property(name="tags", data_type=DataType.TEXT_ARRAY),
            Property(name="scope_type", data_type=DataType.TEXT),
            Property(name="scope_id", data_type=DataType.TEXT),
            Property(name="repo", data_type=DataType.TEXT),
            Property(name="rel_path", data_type=DataType.TEXT),
            Property(name="confidence", data_type=DataType.TEXT),
            Property(name="verification", data_type=DataType.TEXT),
            Property(name="created_at", data_type=DataType.TEXT),
            Property(name="updated_at", data_type=DataType.TEXT),
            Property(name="error_signatures", data_type=DataType.TEXT_ARRAY),
            Property(name="files", data_type=DataType.TEXT_ARRAY),
            Property(name="commands_run", data_type=DataType.TEXT_ARRAY),
        ]

        # Prefer vectorizing text if module exists; fall back to no vectorizer if it doesn't.
        last_err: Optional[Exception] = None
        vectorizers = (
            (Configure.Vectorizer.none(),)
            if self._embedder
            else (Configure.Vectorizer.text2vec_transformers(), Configure.Vectorizer.none())
        )
        for vectorizer in vectorizers:
            try:
                self._client.collections.create(
                    name=self._collection_name,
                    properties=properties,
                    vectorizer_config=vectorizer,
                )
                return
            except Exception as e:
                last_err = e

        raise ConnectionError(f"Failed to create collection '{self._collection_name}': {last_err}")

    def _embed_text(self, text: str) -> Optional[list[float]]:
        embedder = getattr(self, "_embedder", None)
        if not embedder:
            return None
        try:
            vecs = embedder.embed([text])
            return vecs[0] if vecs else None
        except EmbeddingError:
            return None
        except Exception:
            return None

    def add(self, memory: Memory) -> Memory:
        props = memory_to_weaviate_properties(memory)
        vector = self._embed_text(memory.text)
        try:
            self._collection.data.insert(properties=props, uuid=memory.id, vector=vector)
        except Exception as e:
            raise ConnectionError(f"Failed to insert memory into Weaviate: {e}") from e
        return memory

    def get(self, memory_id: str) -> Memory:
        try:
            res = self._collection.query.fetch_object_by_id(
                uuid=memory_id,
                return_properties=True,
            )
        except Exception as e:
            raise ConnectionError(f"Failed to fetch memory from Weaviate: {e}") from e

        obj = getattr(res, "object", None) or res
        if not obj:
            raise NotFoundError(f"Memory not found: {memory_id}")

        props = getattr(obj, "properties", None)
        uuid = getattr(obj, "uuid", None) or getattr(obj, "id", None) or memory_id
        if props is None:
            raise NotFoundError(f"Memory not found: {memory_id}")

        return memory_from_properties(str(uuid), dict(props))

    def edit(self, memory_id: str, update: MemoryUpdate) -> Memory:
        current = self.get(memory_id)
        patch = update.model_dump(exclude_none=True)

        merged = current.model_copy(update=patch)
        merged.updated_at = _now_iso()

        props = memory_to_weaviate_properties(merged)
        vector = self._embed_text(merged.text)
        try:
            self._collection.data.replace(uuid=memory_id, properties=props, vector=vector)
        except Exception as e:
            raise ConnectionError(f"Failed to update memory in Weaviate: {e}") from e
        return merged

    def delete(self, memory_id: str) -> None:
        try:
            ok = self._collection.data.delete_by_id(uuid=memory_id)
        except Exception as e:
            raise ConnectionError(f"Failed to delete memory in Weaviate: {e}") from e
        if ok is False:
            raise NotFoundError(f"Memory not found: {memory_id}")

    def search(self, req: SearchRequest) -> list[SearchHit]:
        # scope_type filter
        filt = wfilters.Filter.by_property("scope_type").equal(req.scope_type)

        # scope_ids OR filter
        if req.scope_ids:
            scope_filters = [wfilters.Filter.by_property("scope_id").equal(s) for s in req.scope_ids]
            filt = wfilters.Filter.all_of([filt, wfilters.Filter.any_of(scope_filters)])

        if req.include_global and req.scope_type != "global":
            global_filt = wfilters.Filter.by_property("scope_type").equal("global")
            filt = wfilters.Filter.any_of([filt, global_filt])

        query_vector = self._embed_text(req.query)

        try:
            qr = None
            if query_vector is not None:
                try:
                    qr = self._collection.query.hybrid(
                        req.query,
                        vector=query_vector,
                        query_properties=["text"],
                        limit=req.top_k,
                        filters=filt,
                        return_metadata=grpc.MetadataQuery(score=True, distance=True),
                        return_properties=True,
                    )
                except Exception:
                    qr = None

            if qr is None:
                try:
                    qr = self._collection.query.hybrid(
                        req.query,
                        query_properties=["text"],
                        limit=req.top_k,
                        filters=filt,
                        return_metadata=grpc.MetadataQuery(score=True, distance=True),
                        return_properties=True,
                    )
                except Exception as e:
                    if not _missing_vectorizer_error(e):
                        raise
                    qr = self._collection.query.bm25(
                        req.query,
                        query_properties=["text"],
                        limit=req.top_k,
                        filters=filt,
                        return_metadata=grpc.MetadataQuery(score=True),
                        return_properties=True,
                    )
        except Exception as e:
            raise ConnectionError(f"Search failed in Weaviate: {e}") from e

        hits: list[SearchHit] = []
        for obj in getattr(qr, "objects", []) or []:
            uuid = str(getattr(obj, "uuid", ""))
            props = dict(getattr(obj, "properties", {}) or {})
            mem = memory_from_properties(uuid, props)

            md = getattr(obj, "metadata", None)
            score = None
            if md is not None:
                score = getattr(md, "score", None)
                if score is None:
                    dist = getattr(md, "distance", None)
                    if dist is not None:
                        try:
                            score = 1.0 / (1.0 + float(dist))
                        except Exception:
                            score = None
            if score is None:
                score = 0.0

            hits.append(SearchHit(id=uuid, score=float(score), memory=mem, source="weaviate"))
        hits.sort(key=lambda h: h.score, reverse=True)
        return hits

    def list(self, filters: ListFilters) -> list[Memory]:
        filt: Optional[wfilters._Filters] = None

        if filters.scope in ("global", "project"):
            filt = wfilters.Filter.by_property("scope_type").equal(filters.scope)

        if filters.repo:
            f_repo = wfilters.Filter.by_property("repo").equal(filters.repo)
            filt = f_repo if filt is None else wfilters.Filter.all_of([filt, f_repo])

        if filters.scope_ids:
            f_scope = wfilters.Filter.any_of([wfilters.Filter.by_property("scope_id").equal(s) for s in filters.scope_ids])
            filt = f_scope if filt is None else wfilters.Filter.all_of([filt, f_scope])

        if filters.mem_type:
            f2 = wfilters.Filter.by_property("mem_type").equal(filters.mem_type.value)
            filt = f2 if filt is None else wfilters.Filter.all_of([filt, f2])

        if filters.tag:
            f3 = wfilters.Filter.by_property("tags").contains_any([filters.tag])
            filt = f3 if filt is None else wfilters.Filter.all_of([filt, f3])

        try:
            qr = self._collection.query.fetch_objects(
                limit=filters.limit,
                filters=filt,
                sort=grpc.Sort.by_update_time(ascending=False),
                return_properties=True,
                return_metadata=grpc.MetadataQuery(creation_time=True, last_update_time=True),
            )
        except Exception as e:
            raise ConnectionError(f"List failed in Weaviate: {e}") from e

        out: list[Memory] = []
        for obj in getattr(qr, "objects", []) or []:
            uuid = str(getattr(obj, "uuid", ""))
            props = dict(getattr(obj, "properties", {}) or {})
            out.append(memory_from_properties(uuid, props))
        return out



from __future__ import annotations

from datetime import datetime, timezone
from enum import Enum
from typing import Any, Optional
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field, field_validator


def _now_dt() -> datetime:
    return datetime.now(tz=timezone.utc).replace(microsecond=0)


def _now_iso() -> str:
    return _now_dt().isoformat()


class MemType(str, Enum):
    finding = "finding"
    recipe = "recipe"
    gotcha = "gotcha"
    preference = "preference"


class ScopeType(str, Enum):
    global_ = "global"
    project = "project"


class Confidence(str, Enum):
    low = "low"
    med = "med"
    high = "high"


class Memory(BaseModel):
    model_config = ConfigDict(validate_assignment=True)

    id: str = Field(default_factory=lambda: str(uuid4()))

    text: str = Field(min_length=1)
    mem_type: MemType
    tags: list[str] = Field(default_factory=list)

    scope_type: ScopeType
    scope_id: str = Field(min_length=1)
    repo: str = Field(default="")
    rel_path: str = Field(default="")

    confidence: Confidence = Field(default=Confidence.med)
    verification: Optional[str] = Field(default=None)

    created_at: datetime = Field(default_factory=_now_dt)
    updated_at: datetime = Field(default_factory=_now_dt)

    error_signatures: Optional[list[str]] = Field(default=None)
    files: Optional[list[str]] = Field(default=None)
    commands_run: Optional[list[str]] = Field(default=None)

    @field_validator("tags")
    @classmethod
    def _dedupe_tags(cls, v: list[str]) -> list[str]:
        cleaned: list[str] = []
        seen: set[str] = set()
        for t in v:
            t2 = t.strip()
            if not t2:
                continue
            if t2 in seen:
                continue
            cleaned.append(t2)
            seen.add(t2)
        return cleaned

    @field_validator("rel_path")
    @classmethod
    def _normalize_rel_path(cls, v: str) -> str:
        v = (v or "").strip()
        if v in (".", "./"):
            return ""
        return v.replace("\\", "/")

    @field_validator("text")
    @classmethod
    def _strip_text(cls, v: str) -> str:
        v = v.strip()
        if not v:
            raise ValueError("text must be non-empty")
        return v

    @field_validator("created_at", "updated_at", mode="before")
    @classmethod
    def _blank_dt_to_now(cls, v: Any) -> Any:
        if isinstance(v, str) and not v.strip():
            return _now_dt()
        return v

    @field_validator("created_at", "updated_at")
    @classmethod
    def _normalize_dt(cls, v: datetime) -> datetime:
        if v.tzinfo is None:
            v = v.replace(tzinfo=timezone.utc)
        else:
            v = v.astimezone(timezone.utc)
        return v.replace(microsecond=0)


class MemoryUpdate(BaseModel):
    text: Optional[str] = None
    mem_type: Optional[MemType] = None
    tags: Optional[list[str]] = None
    confidence: Optional[Confidence] = None
    verification: Optional[str] = None

    @field_validator("text")
    @classmethod
    def _strip_text(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return None
        v = v.strip()
        if not v:
            raise ValueError("text must be non-empty")
        return v


class SearchHit(BaseModel):
    id: str
    score: float
    memory: Memory
    source: str = Field(default="weaviate")  # "weaviate" | "chroma" | "merged"


class ListFilters(BaseModel):
    scope: str = Field(default="project")  # "global" | "project" | "all"
    mem_type: Optional[MemType] = None
    tag: Optional[str] = None
    limit: int = Field(default=20, ge=1, le=200)

    # Optional partitioning helpers (used by CLI for repo-local listing and sync).
    repo: Optional[str] = None
    scope_ids: Optional[list[str]] = None


def memory_to_weaviate_properties(memory: Memory) -> dict[str, Any]:
    props: dict[str, Any] = {
        "text": memory.text,
        "mem_type": memory.mem_type.value,
        "tags": memory.tags,
        "scope_type": memory.scope_type.value,
        "scope_id": memory.scope_id,
        "repo": memory.repo,
        "rel_path": memory.rel_path,
        "confidence": memory.confidence.value,
        "verification": memory.verification,
        "created_at": memory.created_at.isoformat(),
        "updated_at": memory.updated_at.isoformat(),
        "error_signatures": memory.error_signatures,
        "files": memory.files,
        "commands_run": memory.commands_run,
    }
    return {k: v for k, v in props.items() if v is not None}


def memory_from_properties(id: str, props: dict[str, Any]) -> Memory:
    return Memory.model_validate(
        {
            "id": id,
            **props,
        }
    )



from __future__ import annotations

from pathlib import Path
from typing import Any

import pytest

import vibemem.api as api
from vibemem.config import VibememConfig
from vibemem.models import MemType, Memory, ScopeType, SearchHit


def test_search_passes_scope_chain_into_request(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    repo_root = tmp_path / "myrepo"
    (repo_root / ".git").mkdir(parents=True)
    cwd = repo_root / "a" / "b"
    cwd.mkdir(parents=True)

    captured: dict[str, Any] = {}

    class StubStore:
        def __init__(self, cfg: VibememConfig) -> None:
            self._cfg = cfg

        def search(self, req: api.SearchRequest) -> list[SearchHit]:
            captured["req"] = req
            return []

        def close(self) -> None:
            captured["closed"] = True

    monkeypatch.setattr(api, "WeaviateStore", StubStore)

    vm = api.VibeMem(
        cfg=VibememConfig(weaviate_url="http://example:8080"),
        cwd=cwd,
        repo_root=repo_root,
        granularity="cwd",
    )
    result = vm.search("hello", cache="off", include_parents=True)

    req = captured["req"]
    assert req.scope_type == "project"
    assert list(req.scope_ids) == ["myrepo::a/b", "myrepo::a", "myrepo"]
    assert captured["closed"] is True
    assert result.scope_ids == ["myrepo::a/b", "myrepo::a", "myrepo"]
    assert result.scope.scope_id == "myrepo::a/b"


def test_search_falls_back_to_chroma_when_weaviate_unreachable(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    repo_root = tmp_path / "repo"
    (repo_root / ".git").mkdir(parents=True)

    class StubChroma:
        def __init__(self, *, path: Path, enabled: bool) -> None:
            self._enabled = enabled

        def enabled(self) -> bool:
            return self._enabled

        def search(self, req: api.SearchRequest) -> list[SearchHit]:
            mem = Memory(
                text="cached memory",
                mem_type=MemType.finding,
                tags=[],
                scope_type=ScopeType.project,
                scope_id=req.scope_ids[0] if req.scope_ids else "repo",
                repo="repo",
                rel_path="",
            )
            return [SearchHit(id="1", score=0.5, memory=mem, source="chroma")]

    monkeypatch.setattr(api, "ChromaCache", StubChroma)

    vm = api.VibeMem(cfg=VibememConfig(weaviate_url=None), cwd=repo_root, repo_root=repo_root)
    result = vm.search("hello", cache="on")

    assert result.used == "chroma"
    assert [h.id for h in result.hits] == ["1"]
    assert result.scope.scope_type == "project"


def test_scope_info_matches_derived_scope(tmp_path: Path) -> None:
    repo_root = tmp_path / "myrepo"
    (repo_root / ".git").mkdir(parents=True)
    cwd = repo_root / "x"
    cwd.mkdir(parents=True)

    s = api.scope_info(cfg=VibememConfig(weaviate_url="http://example:8080"), cwd=cwd, repo_root=repo_root, granularity="cwd")
    assert s.scope_type == "project"
    assert s.scope_id == "myrepo::x"


def test_add_memory_can_target_global_scope(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    repo_root = tmp_path / "myrepo"
    (repo_root / ".git").mkdir(parents=True)
    cwd = repo_root / "subdir"
    cwd.mkdir(parents=True)

    captured: dict[str, Any] = {}

    class StubStore:
        def __init__(self, cfg: VibememConfig) -> None:
            self._cfg = cfg

        def add(self, memory: Memory) -> Memory:
            captured["memory"] = memory
            return memory

        def close(self) -> None:
            return

    monkeypatch.setattr(api, "WeaviateStore", StubStore)

    vm = api.VibeMem(cfg=VibememConfig(weaviate_url=None), cwd=cwd, repo_root=repo_root)
    created = vm.add_memory(mem_type="recipe", scope="global", text="Use X", tags=["python"], cache="off")

    assert created.scope_type == ScopeType.global_
    assert created.scope_id == "global"
    assert created.repo == "myrepo"
    assert captured["memory"].scope_type == ScopeType.global_


def test_add_memory_allows_explicit_project_scope_id(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, Any] = {}

    class StubStore:
        def __init__(self, cfg: VibememConfig) -> None:
            self._cfg = cfg

        def add(self, memory: Memory) -> Memory:
            captured["memory"] = memory
            return memory

        def close(self) -> None:
            return

    monkeypatch.setattr(api, "WeaviateStore", StubStore)

    vm = api.VibeMem(cfg=VibememConfig(weaviate_url=None))
    created = vm.add_memory(
        mem_type="recipe",
        scope="project",
        scope_id="otherrepo::a/b",
        text="Use X",
        tags="python",
        cache="off",
    )

    assert created.scope_type == ScopeType.project
    assert created.scope_id == "otherrepo::a/b"
    assert created.repo == "otherrepo"
    assert created.rel_path == "a/b"
    assert captured["memory"].repo == "otherrepo"



from __future__ import annotations

from pathlib import Path

import pytest

from vibemem.scope import derive_scope, scope_chain


def test_derive_scope_finds_git_root(tmp_path: Path) -> None:
    repo = tmp_path / "myrepo"
    (repo / ".git").mkdir(parents=True)
    sub = repo / "services" / "api"
    sub.mkdir(parents=True)

    s = derive_scope(sub)

    assert s.repo_root == repo
    assert s.repo_slug == "myrepo"
    assert s.rel_path == "services/api"
    assert s.scope_type == "project"
    assert s.scope_id == "myrepo"


def test_derive_scope_granularity_cwd(tmp_path: Path) -> None:
    repo = tmp_path / "myrepo"
    (repo / ".git").mkdir(parents=True)
    sub = repo / "services" / "api"
    sub.mkdir(parents=True)

    s = derive_scope(sub, granularity="cwd")
    assert s.scope_id == "myrepo::services/api"


def test_derive_scope_granularity_path_n(tmp_path: Path) -> None:
    repo = tmp_path / "myrepo"
    (repo / ".git").mkdir(parents=True)
    sub = repo / "services" / "api" / "v1"
    sub.mkdir(parents=True)

    s1 = derive_scope(sub, granularity="path:1")
    assert s1.scope_id == "myrepo::services"

    s2 = derive_scope(sub, granularity="path:2")
    assert s2.scope_id == "myrepo::services/api"

    s0 = derive_scope(sub, granularity="path:0")
    assert s0.scope_id == "myrepo"


def test_scope_chain_bubbles_up_from_scope_id(tmp_path: Path) -> None:
    repo = tmp_path / "myrepo"
    (repo / ".git").mkdir(parents=True)
    sub = repo / "a" / "b" / "c"
    sub.mkdir(parents=True)

    s = derive_scope(sub, granularity="cwd")
    assert s.scope_id == "myrepo::a/b/c"

    chain = scope_chain(s, include_parents=True)
    assert chain == ["myrepo::a/b/c", "myrepo::a/b", "myrepo::a", "myrepo"]


def test_outside_repo_defaults_to_global(tmp_path: Path) -> None:
    outside = tmp_path / "outside"
    outside.mkdir(parents=True)

    s = derive_scope(outside)
    assert s.scope_type == "global"
    assert s.scope_id == "global"
    assert s.repo_root is None


from __future__ import annotations

import json
import re
from dataclasses import asdict, is_dataclass
from typing import Any, Iterable, Optional, Sequence

from pydantic import BaseModel
from rich.console import Console
from rich.pretty import Pretty


_WORD_RE = re.compile(r"[A-Za-z0-9_]+")


def parse_csv_list(s: Optional[str]) -> list[str]:
    if not s:
        return []
    parts = [p.strip() for p in s.split(",")]
    return [p for p in parts if p]


def uniq(seq: Iterable[str]) -> list[str]:
    out: list[str] = []
    seen: set[str] = set()
    for x in seq:
        if x in seen:
            continue
        seen.add(x)
        out.append(x)
    return out


def normalize_where_paths(paths: Sequence[str]) -> list[str]:
    return [p.replace("\\", "/").strip() for p in paths if p.strip()]


def _to_primitive(obj: Any) -> Any:
    if obj is None:
        return None
    if isinstance(obj, (str, int, float, bool)):
        return obj
    if isinstance(obj, dict):
        return {str(k): _to_primitive(v) for k, v in obj.items()}
    if isinstance(obj, (list, tuple, set)):
        return [_to_primitive(v) for v in obj]
    if isinstance(obj, BaseModel):
        return _to_primitive(obj.model_dump(mode="json"))
    if is_dataclass(obj) and not isinstance(obj, type):
        return _to_primitive(asdict(obj))
    return str(obj)


def emit(obj: Any, *, as_json: bool) -> None:
    if as_json:
        print(json.dumps(_to_primitive(obj), indent=None, separators=(",", ":"), sort_keys=False))
        return

    console = Console()
    console.print(Pretty(_to_primitive(obj), max_depth=6))


def simple_text_score(query: str, text: str) -> float:
    q_words = set(w.lower() for w in _WORD_RE.findall(query))
    t_words = set(w.lower() for w in _WORD_RE.findall(text))
    if not q_words or not t_words:
        return 0.0
    inter = len(q_words & t_words)
    return inter / float(len(q_words))


def dedupe_hits_by_id(hits: Sequence[dict[str, Any]]) -> list[dict[str, Any]]:
    out: list[dict[str, Any]] = []
    seen: set[str] = set()
    for h in hits:
        hid = str(h.get("id", ""))
        if not hid or hid in seen:
            continue
        seen.add(hid)
        out.append(h)
    return out


from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Literal, NoReturn, Optional

import typer

from vibemem.config import (
    VibememConfig,
    load_config,
    read_config_file as read_config_file_data,
    redact_config_for_display,
    write_config_file as write_config_file_data,
)
from vibemem.models import Confidence, ListFilters, MemType, Memory, MemoryUpdate, ScopeType
from vibemem.scope import ScopeInfo, derive_scope, scope_chain
from vibemem.store.base import ConfigError, ConnectionError, NotFoundError, SearchRequest
from vibemem.store.chroma_cache import ChromaCache, chroma_present_on_disk, default_chroma_path
from vibemem.store.weaviate_store import WeaviateStore
from vibemem.util import emit, normalize_where_paths, parse_csv_list, uniq


app = typer.Typer(
    name="vibemem",
    help="VibeMem: manage 'memories' for vibe-coding agents (Weaviate + optional Chroma cache).",
    add_completion=False,
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)

config_app = typer.Typer(
    help="Configuration commands.",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)
app.add_typer(config_app, name="config")


ScopeTypeOpt = Optional[Literal["global", "project"]]
CacheModeOpt = Literal["auto", "on", "off"]


@dataclass(frozen=True)
class Runtime:
    cfg: VibememConfig
    scope: ScopeInfo
    as_json: bool
    config_error: Optional[str] = None


def _parse_scope_type(scope_type: Optional[str]) -> ScopeTypeOpt:
    if scope_type is None:
        return None
    st = scope_type.strip().lower()
    if st not in ("global", "project"):
        raise typer.BadParameter("scope-type must be 'global' or 'project'")
    return st  # type: ignore[return-value]


def _validate_granularity(granularity: str) -> str:
    g = (granularity or "").strip().lower()
    if g in ("repo", "cwd"):
        return g
    if g.startswith("path:"):
        n = g.split(":", 1)[1]
        try:
            if int(n) <= 0:
                raise ValueError
        except ValueError:
            raise typer.BadParameter("granularity 'path:N' requires N to be a positive integer")
        return g
    raise typer.BadParameter("granularity must be 'repo', 'cwd', or 'path:N'")


def _cache_enabled(mode: CacheModeOpt, scope: ScopeInfo) -> tuple[bool, Path]:
    path = default_chroma_path(scope)
    if mode == "off":
        return False, path
    if mode == "on":
        return True, path
    # auto
    return chroma_present_on_disk(path), path


def _die(rt: Runtime, message: str, *, code: int = 1, error_type: str = "error") -> NoReturn:
    emit({"ok": False, "error": {"type": error_type, "message": message}}, as_json=rt.as_json)
    raise typer.Exit(code=code)


def _open_cache_if_enabled(rt: Runtime, mode: CacheModeOpt, scope: ScopeInfo) -> ChromaCache:
    enabled, path = _cache_enabled(mode, scope)
    try:
        return ChromaCache(path=path, enabled=enabled)
    except ModuleNotFoundError:
        if enabled:
            _die(rt, "Chroma cache requested but 'chromadb' is not installed.", code=2, error_type="missing_dependency")
        return ChromaCache(path=path, enabled=False)


def _open_store(rt: Runtime) -> WeaviateStore:
    return WeaviateStore(rt.cfg)


def _open_store_or_die(rt: Runtime) -> WeaviateStore:
    try:
        return WeaviateStore(rt.cfg)
    except ConnectionError as e:
        _die(rt, str(e), code=2, error_type="connection")
        raise


def _apply_specificity_boost(hits: list[dict], scope_ids: list[str]) -> list[dict]:
    if not scope_ids:
        return hits

    idx: dict[str, int] = {sid: i for i, sid in enumerate(scope_ids)}
    denom = max(len(scope_ids) - 1, 1)

    for h in hits:
        mem = h.get("memory", {})
        try:
            base = float(h.get("score", 0.0))
        except Exception:
            base = 0.0

        scope_type = mem.get("scope_type")
        scope_id = mem.get("scope_id")

        boost = 0.0
        if scope_type == "project" and isinstance(scope_id, str) and scope_id in idx:
            specificity = (denom - idx[scope_id]) / float(denom)
            boost = 0.05 * specificity
        elif scope_type == "global":
            boost = 0.01

        h["score"] = base + boost

    hits.sort(key=lambda x: float(x.get("score", 0.0)), reverse=True)
    return hits


@app.callback()
def _main(
    ctx: typer.Context,
    json_output: Annotated[bool, typer.Option("--json/--human", help="Default is --json.")] = True,
    repo_root: Annotated[Optional[Path], typer.Option("--repo-root", help="Override repo root path.")] = None,
    scope_type: Annotated[
        Optional[str], typer.Option("--scope-type", help="Override scope type: global|project.")
    ] = None,
    scope_id: Annotated[Optional[str], typer.Option("--scope-id", help="Override scope id.")] = None,
    granularity: Annotated[str, typer.Option("--granularity", help="repo|cwd|path:N")] = "repo",
) -> None:
    config_error: Optional[str] = None
    try:
        cfg = load_config()
    except Exception as e:
        # Allow `vibemem config ...` commands to run even if the config file is invalid,
        # so users can repair it via the CLI.
        if ctx.invoked_subcommand == "config":
            cfg = VibememConfig()
            config_error = str(e)
        else:
            # ctx.obj not available yet; emit minimal error
            if json_output:
                import json as _json

                print(
                    _json.dumps(
                        {"ok": False, "error": {"type": "config", "message": str(e)}}, separators=(",", ":")
                    )
                )
            else:
                raise typer.BadParameter(str(e))
            raise typer.Exit(code=2)

    scope_type_override = _parse_scope_type(scope_type)
    granularity = _validate_granularity(granularity)

    scope = derive_scope(
        Path.cwd(),
        repo_root_override=repo_root,
        scope_type_override=scope_type_override,
        scope_id_override=scope_id,
        granularity=granularity,
    )
    ctx.obj = Runtime(cfg=cfg, scope=scope, as_json=json_output, config_error=config_error)


@app.command()
def scope(ctx: typer.Context) -> None:
    rt: Runtime = ctx.obj
    enabled_auto, chroma_path = _cache_enabled(rt.cfg.cache_mode, rt.scope)
    emit(
        {
            "cwd": str(rt.scope.cwd),
            "repo_root": str(rt.scope.repo_root) if rt.scope.repo_root else None,
            "repo_slug": rt.scope.repo_slug,
            "rel_path": rt.scope.rel_path,
            "scope_type": rt.scope.scope_type,
            "scope_id": rt.scope.scope_id,
            "cache_mode": rt.cfg.cache_mode,
            "chroma_path": str(chroma_path),
            "chroma_present": chroma_present_on_disk(chroma_path),
            "chroma_enabled_auto": enabled_auto,
        },
        as_json=rt.as_json,
    )


@config_app.command("show")
def config_show(ctx: typer.Context) -> None:
    rt: Runtime = ctx.obj
    out = redact_config_for_display(rt.cfg)
    if rt.config_error:
        out["config_error"] = rt.config_error
    emit(out, as_json=rt.as_json)


@config_app.command("path")
def config_path(ctx: typer.Context) -> None:
    rt: Runtime = ctx.obj
    emit({"config_file": str(VibememConfig.config_file())}, as_json=rt.as_json)


@config_app.command("init")
def config_init(
    ctx: typer.Context,
    force: Annotated[bool, typer.Option("--force", help="Overwrite existing config file.")] = False,
) -> None:
    rt: Runtime = ctx.obj
    path = VibememConfig.config_file()
    if path.exists() and not force:
        _die(rt, f"Config file already exists: {path}", code=2, error_type="config")

    write_config_file_data({}, path=path)
    emit({"ok": True, "config_file": str(path)}, as_json=rt.as_json)


@config_app.command("set")
def config_set(
    ctx: typer.Context,
    weaviate_url: Annotated[Optional[str], typer.Option("--weaviate-url")] = None,
    weaviate_grpc_url: Annotated[Optional[str], typer.Option("--weaviate-grpc-url")] = None,
    weaviate_api_key: Annotated[Optional[str], typer.Option("--weaviate-api-key")] = None,
    clear_weaviate_api_key: Annotated[
        bool, typer.Option("--clear-weaviate-api-key", help="Remove the API key from the config file.")
    ] = False,
    embedding_host: Annotated[Optional[str], typer.Option("--embedding-host")] = None,
    embedding_port: Annotated[Optional[int], typer.Option("--embedding-port")] = None,
    embedding_model: Annotated[Optional[str], typer.Option("--embedding-model")] = None,
    weaviate_collection: Annotated[Optional[str], typer.Option("--weaviate-collection")] = None,
    cache_mode: Annotated[Optional[CacheModeOpt], typer.Option("--cache-mode", help="auto|on|off")] = None,
    request_timeout_s: Annotated[Optional[float], typer.Option("--request-timeout-s")] = None,
    force: Annotated[
        bool,
        typer.Option("--force", help="Overwrite even if the existing config file contains invalid JSON."),
    ] = False,
) -> None:
    rt: Runtime = ctx.obj
    path = VibememConfig.config_file()

    if (
        weaviate_url is None
        and weaviate_grpc_url is None
        and weaviate_api_key is None
        and not clear_weaviate_api_key
        and embedding_host is None
        and embedding_port is None
        and embedding_model is None
        and weaviate_collection is None
        and cache_mode is None
        and request_timeout_s is None
    ):
        raise typer.BadParameter("No settings provided. Pass one or more --weaviate-* / --cache-mode / --request-timeout-s.")

    try:
        current = read_config_file_data(path=path)
    except ConfigError as e:
        if not force:
            _die(rt, str(e), code=2, error_type="config")
        current = {}

    updated = dict(current)

    if weaviate_url is not None:
        updated["weaviate_url"] = weaviate_url
    if weaviate_grpc_url is not None:
        updated["weaviate_grpc_url"] = weaviate_grpc_url
    if embedding_host is not None:
        updated["embedding_host"] = embedding_host
    if embedding_port is not None:
        updated["embedding_port"] = embedding_port
    if embedding_model is not None:
        updated["embedding_model"] = embedding_model
    if weaviate_collection is not None:
        updated["weaviate_collection"] = weaviate_collection
    if cache_mode is not None:
        updated["cache_mode"] = cache_mode
    if request_timeout_s is not None:
        updated["request_timeout_s"] = request_timeout_s

    if clear_weaviate_api_key:
        updated.pop("weaviate_api_key", None)
    elif weaviate_api_key is not None:
        updated["weaviate_api_key"] = weaviate_api_key

    # Validate before writing. Keep the raw dict (so secrets stay as strings on disk).
    try:
        VibememConfig.model_validate(updated)
    except Exception as e:
        _die(rt, f"Invalid config: {e}", code=2, error_type="config")

    write_config_file_data(updated, path=path)

    # Show effective config (env vars override file values).
    emit({"ok": True, **redact_config_for_display(load_config())}, as_json=rt.as_json)


@config_app.command("unset")
def config_unset(
    ctx: typer.Context,
    keys: Annotated[list[str], typer.Argument(help="One or more config keys to remove.")],
    force: Annotated[
        bool,
        typer.Option("--force", help="Overwrite even if the existing config file contains invalid JSON."),
    ] = False,
) -> None:
    rt: Runtime = ctx.obj
    path = VibememConfig.config_file()

    if not keys:
        raise typer.BadParameter("Provide one or more keys to unset.")

    try:
        current = read_config_file_data(path=path)
    except ConfigError as e:
        if not force:
            _die(rt, str(e), code=2, error_type="config")
        current = {}

    updated = dict(current)
    for k in keys:
        updated.pop(k, None)

    try:
        VibememConfig.model_validate(updated)
    except Exception as e:
        _die(rt, f"Invalid config: {e}", code=2, error_type="config")

    write_config_file_data(updated, path=path)
    emit({"ok": True, **redact_config_for_display(load_config())}, as_json=rt.as_json)


@app.command()
def search(
    ctx: typer.Context,
    query: Annotated[str, typer.Argument(help="Search query text.")],
    top: Annotated[int, typer.Option("--top", help="Max results.")] = 8,
    include_global: Annotated[bool, typer.Option("--include-global/--no-include-global")] = True,
    include_parents: Annotated[bool, typer.Option("--include-parents/--no-include-parents")] = True,
    cache: Annotated[CacheModeOpt, typer.Option("--cache", help="auto|on|off")] = "auto",
) -> None:
    rt: Runtime = ctx.obj

    scope_ids: list[str] = []
    if rt.scope.scope_type == "project":
        scope_ids = scope_chain(rt.scope, include_parents=include_parents)
    else:
        include_global = False

    cache_mode = cache if cache != "auto" else rt.cfg.cache_mode
    chroma = _open_cache_if_enabled(rt, cache_mode, rt.scope)

    req = SearchRequest(
        query=query,
        scope_type=rt.scope.scope_type,
        scope_ids=scope_ids,
        include_global=include_global,
        top_k=top,
    )

    hits: list[dict] = []
    used: str = "weaviate"

    try:
        store = _open_store(rt)
        try:
            whits = store.search(req)
        finally:
            store.close()

        hits = [h.model_dump(mode="json") for h in whits]
    except ConnectionError as e:
        if chroma.enabled():
            used = "chroma"
            chits = chroma.search(req)
            hits = [h.model_dump(mode="json") for h in chits]
        else:
            _die(rt, str(e), code=2, error_type="connection")

    hits = _apply_specificity_boost(hits, scope_ids)
    deduped: list[dict] = []
    seen: set[str] = set()
    for h in hits:
        hid = str(h.get("id", ""))
        if not hid or hid in seen:
            continue
        seen.add(hid)
        deduped.append(h)

    emit(
        {
            "query": query,
            "scope_type": rt.scope.scope_type,
            "scope_ids": scope_ids,
            "include_global": include_global,
            "include_parents": include_parents,
            "used": used,
            "hits": deduped[:top],
        },
        as_json=rt.as_json,
    )


@app.command()
def add(
    ctx: typer.Context,
    mem_type: Annotated[MemType, typer.Option("--type", help="finding|recipe|gotcha|preference")],
    text: Annotated[str, typer.Option("--text", help="Memory content.")],
    tags: Annotated[Optional[str], typer.Option("--tags", help="Comma-separated tags.")] = None,
    confidence: Annotated[Confidence, typer.Option("--confidence")] = Confidence.med,
    verification: Annotated[Optional[str], typer.Option("--verification")] = None,
    errors: Annotated[list[str], typer.Option("--error", help="Repeatable error signature.")] = [],
    files: Annotated[list[str], typer.Option("--file", help="Repeatable file path.")] = [],
    cmds: Annotated[list[str], typer.Option("--cmd", help="Repeatable command run.")] = [],
    cache: Annotated[CacheModeOpt, typer.Option("--cache", help="auto|on|off")] = "auto",
) -> None:
    rt: Runtime = ctx.obj
    cache_mode = cache if cache != "auto" else rt.cfg.cache_mode
    chroma = _open_cache_if_enabled(rt, cache_mode, rt.scope)

    m = Memory(
        text=text,
        mem_type=mem_type,
        tags=uniq(parse_csv_list(tags)),
        scope_type=ScopeType(rt.scope.scope_type),
        scope_id=rt.scope.scope_id,
        repo=rt.scope.repo_slug,
        rel_path=rt.scope.rel_path,
        confidence=confidence,
        verification=verification,
        error_signatures=errors or None,
        files=normalize_where_paths(files) or None,
        commands_run=cmds or None,
    )

    try:
        store = _open_store_or_die(rt)
        try:
            created = store.add(m)
        finally:
            store.close()
    except ConnectionError as e:
        _die(rt, str(e), code=2, error_type="connection")

    if chroma.enabled():
        chroma.upsert(created)

    emit(created, as_json=rt.as_json)


@app.command()
def edit(
    ctx: typer.Context,
    memory_id: Annotated[str, typer.Argument(help="Memory UUID.")],
    text: Annotated[Optional[str], typer.Option("--text")] = None,
    mem_type: Annotated[Optional[MemType], typer.Option("--type")] = None,
    tags: Annotated[Optional[str], typer.Option("--tags")] = None,
    confidence: Annotated[Optional[Confidence], typer.Option("--confidence")] = None,
    verification: Annotated[Optional[str], typer.Option("--verification")] = None,
    cache: Annotated[CacheModeOpt, typer.Option("--cache", help="auto|on|off")] = "auto",
) -> None:
    rt: Runtime = ctx.obj
    cache_mode = cache if cache != "auto" else rt.cfg.cache_mode
    chroma = _open_cache_if_enabled(rt, cache_mode, rt.scope)

    upd = MemoryUpdate(
        text=text,
        mem_type=mem_type,
        tags=uniq(parse_csv_list(tags)) if tags is not None else None,
        confidence=confidence,
        verification=verification,
    )

    try:
        store = _open_store_or_die(rt)
        try:
            updated = store.edit(memory_id, upd)
        finally:
            store.close()
    except NotFoundError as e:
        _die(rt, str(e), code=3, error_type="not_found")
    except ConnectionError as e:
        _die(rt, str(e), code=2, error_type="connection")

    if chroma.enabled():
        chroma.upsert(updated)

    emit(updated, as_json=rt.as_json)


@app.command("rm")
def rm_cmd(
    ctx: typer.Context,
    memory_id: Annotated[str, typer.Argument(help="Memory UUID.")],
    cache: Annotated[CacheModeOpt, typer.Option("--cache", help="auto|on|off")] = "auto",
) -> None:
    rt: Runtime = ctx.obj
    cache_mode = cache if cache != "auto" else rt.cfg.cache_mode
    chroma = _open_cache_if_enabled(rt, cache_mode, rt.scope)

    try:
        store = _open_store_or_die(rt)
        try:
            store.delete(memory_id)
        finally:
            store.close()
    except NotFoundError as e:
        _die(rt, str(e), code=3, error_type="not_found")
    except ConnectionError as e:
        _die(rt, str(e), code=2, error_type="connection")

    if chroma.enabled():
        chroma.delete(memory_id)

    emit({"deleted": memory_id}, as_json=rt.as_json)


@app.command("list")
def list_cmd(
    ctx: typer.Context,
    scope: Annotated[str, typer.Option("--scope", help="global|project|all")] = "project",
    mem_type: Annotated[Optional[MemType], typer.Option("--type")] = None,
    tag: Annotated[Optional[str], typer.Option("--tag")] = None,
    limit: Annotated[int, typer.Option("--limit")] = 20,
) -> None:
    rt: Runtime = ctx.obj
    scope = scope.strip().lower()
    if scope not in ("global", "project", "all"):
        raise typer.BadParameter("scope must be global|project|all")

    repo_filter: Optional[str] = None
    if scope == "project" and rt.scope.repo_slug:
        repo_filter = rt.scope.repo_slug

    filters = ListFilters(scope=scope, mem_type=mem_type, tag=tag, limit=limit, repo=repo_filter, scope_ids=None)

    try:
        store = _open_store_or_die(rt)
        try:
            memories = store.list(filters)
        finally:
            store.close()
    except ConnectionError as e:
        _die(rt, str(e), code=2, error_type="connection")

    emit(
        {"scope": scope, "repo": repo_filter, "count": len(memories), "memories": [m.model_dump(mode="json") for m in memories]},
        as_json=rt.as_json,
    )


@app.command()
def sync(
    ctx: typer.Context,
    pull: Annotated[bool, typer.Option("--pull", help="Pull from Weaviate into local Chroma cache.")] = False,
    limit: Annotated[int, typer.Option("--limit")] = 200,
) -> None:
    rt: Runtime = ctx.obj
    if not pull:
        emit({"todo": "push/offline queue not implemented yet", "hint": "use --pull"}, as_json=rt.as_json)
        raise typer.Exit(code=2)

    if rt.scope.scope_type != "project":
        raise typer.BadParameter("sync --pull requires project scope (run inside a repo or set --scope-type project).")

    chroma = _open_cache_if_enabled(rt, "on", rt.scope)
    if not chroma.enabled():
        raise typer.Exit(code=2)

    scope_ids = scope_chain(rt.scope, include_parents=True)

    try:
        store = _open_store_or_die(rt)
        try:
            memories = store.list(
                ListFilters(
                    scope="project",
                    mem_type=None,
                    tag=None,
                    limit=limit,
                    repo=rt.scope.repo_slug,
                    scope_ids=scope_ids,
                )
            )
        finally:
            store.close()
    except ConnectionError as e:
        _die(rt, str(e), code=2, error_type="connection")

    chroma.rebuild(memories)

    emit(
        {
            "pulled": len(memories),
            "scope_ids": scope_ids,
            "chroma_path": str(default_chroma_path(rt.scope)),
            "chroma_count": chroma.count(),
        },
        as_json=rt.as_json,
    )


# Typer/Click errors (BadParameter, etc.) are handled by Typer itself.
# For domain errors we emit JSON and exit non-zero inside command handlers.


if __name__ == "__main__":
    app()



from __future__ import annotations

import importlib
import json
import hashlib
import math
import re
from pathlib import Path
from typing import Any, Optional, Sequence

from vibemem.models import Memory, SearchHit, memory_from_properties
from vibemem.scope import ScopeInfo
from vibemem.store.base import MemoryCache, SearchRequest
from vibemem.util import simple_text_score


_WORD_RE = re.compile(r"[A-Za-z0-9_]+")


def default_chroma_path(scope: ScopeInfo) -> Path:
    base = Path.home() / ".vibemem" / "chroma"
    if scope.repo_root:
        try:
            root = scope.repo_root.resolve()
        except Exception:
            root = scope.repo_root
        digest = hashlib.sha1(str(root).encode("utf-8")).hexdigest()[:10]
        slug = scope.repo_slug or "repo"
        return base / f"{slug}-{digest}"
    return base / "global"


def chroma_present_on_disk(path: Path) -> bool:
    if not path.exists():
        return False
    if path.is_file():
        return False
    try:
        return any(path.iterdir())
    except Exception:
        return False


def _hash_embedding(text: str, *, dim: int = 256) -> list[float]:
    # Deterministic, dependency-free embedding for offline cache usage.
    # Not meant to be SOTA; just stable and "good enough" for small local caches.
    vec = [0.0] * dim
    tokens = [t.lower() for t in _WORD_RE.findall(text)]
    if not tokens:
        return vec

    for t in tokens:
        h = 2166136261
        for ch in t:
            h ^= ord(ch)
            h *= 16777619
            h &= 0xFFFFFFFF
        idx = h % dim
        sign = -1.0 if (h & 1) else 1.0
        vec[idx] += sign * 1.0

    norm = math.sqrt(sum(v * v for v in vec)) or 1.0
    return [v / norm for v in vec]


def _memory_to_chroma_metadata(memory: Memory) -> dict[str, Any]:
    return {
        "mem_type": memory.mem_type.value,
        "tags_csv": ",".join(memory.tags),
        "scope_type": memory.scope_type.value,
        "scope_id": memory.scope_id,
        "repo": memory.repo,
        "rel_path": memory.rel_path,
        "confidence": memory.confidence.value,
        "verification": memory.verification or "",
        "created_at": memory.created_at.isoformat(),
        "updated_at": memory.updated_at.isoformat(),
        "error_signatures_json": json.dumps(memory.error_signatures or []),
        "files_json": json.dumps(memory.files or []),
        "commands_run_json": json.dumps(memory.commands_run or []),
    }


def _memory_from_chroma(id_: str, document: str, metadata: dict[str, Any]) -> Memory:
    def _load_json_list(key: str) -> Optional[list[str]]:
        raw = metadata.get(key)
        if not raw:
            return None
        try:
            parsed = json.loads(raw)
            if isinstance(parsed, list):
                return [str(x) for x in parsed]
        except Exception:
            return None
        return None

    props: dict[str, Any] = {
        "text": document,
        "mem_type": metadata.get("mem_type", "finding"),
        "tags": [t for t in str(metadata.get("tags_csv", "")).split(",") if t],
        "scope_type": metadata.get("scope_type", "project"),
        "scope_id": metadata.get("scope_id", ""),
        "repo": metadata.get("repo", ""),
        "rel_path": metadata.get("rel_path", ""),
        "confidence": metadata.get("confidence", "med"),
        "verification": (metadata.get("verification") or None),
        "error_signatures": _load_json_list("error_signatures_json"),
        "files": _load_json_list("files_json"),
        "commands_run": _load_json_list("commands_run_json"),
    }

    created_at = metadata.get("created_at")
    if created_at:
        props["created_at"] = created_at
    updated_at = metadata.get("updated_at")
    if updated_at:
        props["updated_at"] = updated_at

    return memory_from_properties(id_, props)


class ChromaCache(MemoryCache):
    def __init__(self, *, path: Path, enabled: bool) -> None:
        self._path = path
        self._enabled = enabled
        self._client = None
        self._collection = None

        if not self._enabled:
            return

        self._path.mkdir(parents=True, exist_ok=True)

        chromadb = importlib.import_module("chromadb")
        self._client = chromadb.PersistentClient(path=str(self._path))
        self._collection = self._client.get_or_create_collection(name="VibeMemMemory")

    def enabled(self) -> bool:
        return self._enabled and self._collection is not None

    def upsert(self, memory: Memory) -> None:
        if not self.enabled():
            return
        assert self._collection is not None

        self._collection.upsert(
            ids=[memory.id],
            documents=[memory.text],
            metadatas=[_memory_to_chroma_metadata(memory)],
            embeddings=[_hash_embedding(memory.text)],
        )

    def delete(self, memory_id: str) -> None:
        if not self.enabled():
            return
        assert self._collection is not None
        try:
            self._collection.delete(ids=[memory_id])
        except Exception:
            return

    def search(self, req: SearchRequest) -> list[SearchHit]:
        if not self.enabled():
            return []
        assert self._collection is not None
        collection = self._collection

        def _run_query(where: dict[str, Any]) -> list[SearchHit]:
            # Try vector query first; fall back to scan+score if Chroma's where operators differ.
            try:
                res = collection.query(
                    query_embeddings=[_hash_embedding(req.query)],
                    n_results=req.top_k,
                    where=where,
                    include=["documents", "metadatas", "distances"],
                )
                ids = (res.get("ids") or [[]])[0]
                docs = (res.get("documents") or [[]])[0]
                metas = (res.get("metadatas") or [[]])[0]
                dists = (res.get("distances") or [[]])[0]

                hits: list[SearchHit] = []
                for id_, doc, meta, dist in zip(ids, docs, metas, dists):
                    mem = _memory_from_chroma(str(id_), str(doc), dict(meta or {}))
                    score = 1.0 / (1.0 + float(dist)) if dist is not None else 0.0
                    hits.append(SearchHit(id=str(id_), score=float(score), memory=mem, source="chroma"))
                return hits
            except Exception:
                pass

            try:
                res2 = collection.get(include=["documents", "metadatas", "ids"])
            except Exception:
                return []

            ids2 = res2.get("ids") or []
            docs2 = res2.get("documents") or []
            metas2 = res2.get("metadatas") or []

            allowed_scope_type = where.get("scope_type")
            allowed_scope_ids: Optional[set[str]] = None
            if isinstance(where.get("scope_id"), dict) and "$in" in where["scope_id"]:
                allowed_scope_ids = set(str(x) for x in where["scope_id"]["$in"])

            hits2: list[SearchHit] = []
            for id_, doc, meta in zip(ids2, docs2, metas2):
                meta_d = dict(meta or {})
                if allowed_scope_type is not None and meta_d.get("scope_type") != allowed_scope_type:
                    continue
                if allowed_scope_ids is not None and str(meta_d.get("scope_id", "")) not in allowed_scope_ids:
                    continue
                mem = _memory_from_chroma(str(id_), str(doc), meta_d)
                score = simple_text_score(req.query, mem.text)
                hits2.append(SearchHit(id=str(id_), score=float(score), memory=mem, source="chroma"))
            return hits2

        where_project: dict[str, Any] = {"scope_type": req.scope_type}
        if req.scope_ids:
            where_project["scope_id"] = {"$in": list(req.scope_ids)}

        hits = _run_query(where_project)

        if req.include_global and req.scope_type != "global":
            hits.extend(_run_query({"scope_type": "global"}))

        # De-dupe + rank
        best: dict[str, SearchHit] = {}
        for h in hits:
            if h.id not in best or h.score > best[h.id].score:
                best[h.id] = h

        out = sorted(best.values(), key=lambda h: h.score, reverse=True)
        return out[: req.top_k]

    def rebuild(self, memories: Sequence[Memory]) -> None:
        if not self.enabled():
            return
        assert self._client is not None

        try:
            self._client.delete_collection(name="VibeMemMemory")
        except Exception:
            pass

        self._collection = self._client.get_or_create_collection(name="VibeMemMemory")

        for m in memories:
            self.upsert(m)

    def count(self) -> Optional[int]:
        if not self.enabled():
            return None
        assert self._collection is not None
        try:
            return int(self._collection.count())
        except Exception:
            return None



from __future__ import annotations

from vibemem.api import (
    VibeMem,
    add_memory,
    delete_memory,
    edit_memory,
    get_memory,
    list_memories,
    search,
    search_hits,
    search_memories,
    scope_info,
    sync_pull,
)

__all__ = [
    "__version__",
    "VibeMem",
    "search",
    "search_hits",
    "search_memories",
    "scope_info",
    "add_memory",
    "edit_memory",
    "delete_memory",
    "get_memory",
    "list_memories",
    "sync_pull",
]

__version__ = "0.2.5"



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Literal, Optional

from pydantic import BaseModel, Field, SecretStr

from vibemem.store.base import ConfigError


CacheMode = Literal["auto", "on", "off"]


class VibememConfig(BaseModel):
    weaviate_url: Optional[str] = Field(default=None)
    weaviate_grpc_url: Optional[str] = Field(default=None)
    weaviate_api_key: Optional[SecretStr] = Field(default=None)

    # Optional client-side embeddings (for Weaviate classes with vectorizer = none).
    embedding_host: Optional[str] = Field(default=None)
    embedding_port: Optional[int] = Field(default=None, ge=1, le=65535)
    embedding_model: Optional[str] = Field(default=None)

    weaviate_collection: str = Field(default="VibeMemMemory")
    cache_mode: CacheMode = Field(default="auto")

    request_timeout_s: float = Field(default=10.0, ge=1.0, le=120.0)

    @staticmethod
    def config_dir() -> Path:
        path = Path.home() / ".vibemem"
        path.mkdir(parents=True, exist_ok=True)
        return path

    @staticmethod
    def config_file() -> Path:
        return VibememConfig.config_dir() / "config"


def _read_json_file(path: Path) -> dict:
    if not path.exists():
        return {}
    try:
        data = json.loads(path.read_text(encoding="utf-8"))
    except json.JSONDecodeError as e:
        raise ConfigError(
            f"Invalid JSON in config file {path} (line {e.lineno} column {e.colno}): {e.msg}"
        ) from e
    except OSError as e:
        raise ConfigError(f"Failed to read config file {path}: {e}") from e

    if not isinstance(data, dict):
        raise ConfigError(f"Config file {path} must contain a JSON object at the top level.")
    return data


def read_config_file(*, path: Optional[Path] = None) -> dict:
    return _read_json_file(path or VibememConfig.config_file())


def write_config_file(data: dict, *, path: Optional[Path] = None) -> Path:
    out_path = path or VibememConfig.config_file()
    out_path.parent.mkdir(parents=True, exist_ok=True)

    tmp_path = out_path.with_name(out_path.name + ".tmp")
    payload = json.dumps(data, indent=2, sort_keys=True) + "\n"

    try:
        tmp_path.write_text(payload, encoding="utf-8")
        os.replace(tmp_path, out_path)
    except OSError as e:
        raise ConfigError(f"Failed to write config file {out_path}: {e}") from e
    finally:
        try:
            if tmp_path.exists():
                tmp_path.unlink()
        except OSError:
            pass

    if os.name != "nt":
        try:
            os.chmod(out_path, 0o600)
        except OSError:
            pass

    return out_path


def load_config() -> VibememConfig:
    file_data = _read_json_file(VibememConfig.config_file())

    env_data: dict = {}
    if os.getenv("VIBEMEM_WEAVIATE_URL"):
        env_data["weaviate_url"] = os.getenv("VIBEMEM_WEAVIATE_URL")
    if os.getenv("VIBEMEM_WEAVIATE_GRPC_URL"):
        env_data["weaviate_grpc_url"] = os.getenv("VIBEMEM_WEAVIATE_GRPC_URL")
    if os.getenv("VIBEMEM_WEAVIATE_API_KEY"):
        env_data["weaviate_api_key"] = os.getenv("VIBEMEM_WEAVIATE_API_KEY")
    if os.getenv("VIBEMEM_WEAVIATE_COLLECTION"):
        env_data["weaviate_collection"] = os.getenv("VIBEMEM_WEAVIATE_COLLECTION")
    if os.getenv("VIBEMEM_CACHE_MODE"):
        env_data["cache_mode"] = os.getenv("VIBEMEM_CACHE_MODE")

    if os.getenv("VIBEMEM_EMBEDDING_HOST"):
        env_data["embedding_host"] = os.getenv("VIBEMEM_EMBEDDING_HOST")
    if os.getenv("VIBEMEM_EMBEDDING_PORT"):
        env_data["embedding_port"] = os.getenv("VIBEMEM_EMBEDDING_PORT")
    if os.getenv("VIBEMEM_EMBEDDING_MODEL"):
        env_data["embedding_model"] = os.getenv("VIBEMEM_EMBEDDING_MODEL")

    merged = dict(file_data)
    merged.update(env_data)
    return VibememConfig.model_validate(merged)


def redact_config_for_display(cfg: VibememConfig) -> dict:
    return {
        "weaviate_url": cfg.weaviate_url,
        "weaviate_grpc_url": cfg.weaviate_grpc_url,
        "weaviate_api_key": "***redacted***" if cfg.weaviate_api_key else None,
        "embedding_host": cfg.embedding_host,
        "embedding_port": cfg.embedding_port,
        "embedding_model": cfg.embedding_model,
        "weaviate_collection": cfg.weaviate_collection,
        "cache_mode": cfg.cache_mode,
        "request_timeout_s": cfg.request_timeout_s,
        "config_file": str(cfg.config_file()),
    }



from __future__ import annotations

import json
import socket
from dataclasses import dataclass
from typing import Any, Optional, Sequence
from urllib.parse import urlparse
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen


class EmbeddingError(RuntimeError):
    pass


@dataclass(frozen=True)
class EmbeddingHTTPError(EmbeddingError):
    url: str
    status_code: int
    body: str

    def __str__(self) -> str:  # pragma: no cover
        msg = f"Embedding request failed ({self.status_code}) at {self.url}"
        if self.body:
            msg += f": {self.body}"
        return msg


def _post_json(url: str, payload: Any, *, timeout_s: float) -> Any:
    body = json.dumps(payload).encode("utf-8")
    req = Request(
        url,
        data=body,
        headers={"Content-Type": "application/json", "Accept": "application/json"},
        method="POST",
    )

    try:
        with urlopen(req, timeout=timeout_s) as resp:
            raw = resp.read()
    except HTTPError as e:
        try:
            err_body = e.read().decode("utf-8", errors="replace")
        except Exception:
            err_body = ""
        raise EmbeddingHTTPError(url=url, status_code=int(getattr(e, "code", 0) or 0), body=err_body) from e
    except (URLError, socket.timeout) as e:
        raise EmbeddingError(f"Embedding request failed at {url}: {e}") from e

    try:
        return json.loads(raw.decode("utf-8", errors="replace"))
    except json.JSONDecodeError as e:
        preview = raw[:200].decode("utf-8", errors="replace")
        raise EmbeddingError(f"Invalid JSON from embedding server at {url}: {preview}") from e


def _as_vector(obj: Any) -> list[float]:
    if not isinstance(obj, list):
        raise EmbeddingError("Embedding is not a list.")
    try:
        return [float(x) for x in obj]
    except Exception as e:
        raise EmbeddingError("Embedding contains non-numeric values.") from e


def _parse_openai_embeddings(resp: Any, *, expected: int) -> list[list[float]]:
    if not isinstance(resp, dict):
        raise EmbeddingError("OpenAI embeddings response is not a JSON object.")

    data = resp.get("data")
    if not isinstance(data, list):
        raise EmbeddingError("OpenAI embeddings response missing 'data' list.")

    # Prefer explicit indices if present; fall back to positional order.
    by_index: dict[int, list[float]] = {}
    positional: list[list[float]] = []
    for item in data:
        if not isinstance(item, dict):
            continue
        emb = _as_vector(item.get("embedding"))
        idx = item.get("index")
        if isinstance(idx, int):
            by_index[idx] = emb
        else:
            positional.append(emb)

    if by_index:
        out = [by_index[i] for i in sorted(by_index.keys())]
    else:
        out = positional

    if expected and len(out) != expected:
        raise EmbeddingError(f"Expected {expected} embeddings, got {len(out)}.")
    return out


def _parse_tei_embed(resp: Any, *, expected: int) -> list[list[float]]:
    # TEI returns either a single vector (list[float]) or list[list[float]].
    if isinstance(resp, list) and resp and all(isinstance(x, (int, float)) for x in resp):
        out = [_as_vector(resp)]
    elif isinstance(resp, list) and all(isinstance(row, list) for row in resp):
        out = [_as_vector(row) for row in resp]
    elif isinstance(resp, dict) and "embeddings" in resp:
        out = _parse_tei_embed(resp.get("embeddings"), expected=expected)
    else:
        raise EmbeddingError("Unsupported embedding response format.")

    if expected and len(out) != expected:
        raise EmbeddingError(f"Expected {expected} embeddings, got {len(out)}.")
    return out


@dataclass(frozen=True)
class EmbeddingClient:
    host: str
    port: int
    model: Optional[str]
    timeout_s: float = 10.0

    @property
    def base_url(self) -> str:
        host = self.host.strip()
        if "://" in host:
            parsed = urlparse(host)
            scheme = parsed.scheme or "http"
            hostname = parsed.hostname or ""
            if not hostname:
                raise EmbeddingError(f"Invalid embedding host URL: {host}")
            port = parsed.port or self.port
            path = (parsed.path or "").rstrip("/")
            base = f"{scheme}://{hostname}:{port}"
            return (base + path).rstrip("/")

        host = host.rstrip("/")
        if host.count(":") == 1 and host.rsplit(":", 1)[1].isdigit():
            return f"http://{host}"
        return f"http://{host}:{self.port}"

    def embed(self, texts: Sequence[str]) -> list[list[float]]:
        items = [str(t) for t in texts]
        if not items:
            return []

        # Try OpenAI-compatible endpoint first when a model is provided.
        if self.model:
            try:
                resp = _post_json(
                    f"{self.base_url}/v1/embeddings",
                    {"model": self.model, "input": items},
                    timeout_s=self.timeout_s,
                )
                return _parse_openai_embeddings(resp, expected=len(items))
            except EmbeddingHTTPError as e:
                if e.status_code not in (404, 405):
                    raise
            except EmbeddingError:
                # Fall back to TEI endpoint.
                pass

        resp2 = _post_json(f"{self.base_url}/embed", {"inputs": items}, timeout_s=self.timeout_s)
        return _parse_tei_embed(resp2, expected=len(items))



from __future__ import annotations

from typing import Any

from vibemem.store.base import SearchRequest
from vibemem.store.chroma_cache import ChromaCache


class StubCollection:
    def __init__(self, *, ids: list[str], documents: list[str], metadatas: list[dict[str, Any]]) -> None:
        self._ids = ids
        self._documents = documents
        self._metadatas = metadatas
        self.query_calls = 0
        self.get_calls = 0

    def query(self, *args: Any, **kwargs: Any) -> Any:
        self.query_calls += 1
        raise RuntimeError("query failed")

    def get(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        self.get_calls += 1
        return {"ids": list(self._ids), "documents": list(self._documents), "metadatas": list(self._metadatas)}


def _meta(*, scope_type: str, scope_id: str) -> dict[str, Any]:
    return {
        "mem_type": "finding",
        "tags_csv": "",
        "scope_type": scope_type,
        "scope_id": scope_id,
        "repo": "repo",
        "rel_path": "",
        "confidence": "med",
        "verification": "",
        "created_at": "2025-01-01T00:00:00+00:00",
        "updated_at": "2025-01-01T00:00:00+00:00",
        "error_signatures_json": "[]",
        "files_json": "[]",
        "commands_run_json": "[]",
    }


def test_chroma_search_falls_back_to_scan_when_query_fails() -> None:
    collection = StubCollection(
        ids=["1", "2", "3"],
        documents=["foo bar", "foo baz", "foo foo"],
        metadatas=[
            _meta(scope_type="project", scope_id="repo::a"),
            _meta(scope_type="project", scope_id="repo::other"),
            _meta(scope_type="global", scope_id="global"),
        ],
    )

    cache = object.__new__(ChromaCache)
    cache._enabled = True
    cache._collection = collection

    req = SearchRequest(
        query="foo bar",
        scope_type="project",
        scope_ids=["repo::a", "repo"],
        include_global=True,
        top_k=10,
    )
    hits = cache.search(req)

    assert collection.query_calls >= 1
    assert collection.get_calls >= 1
    assert [h.id for h in hits] == ["1", "3"]




from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Protocol, Sequence

from vibemem.models import ListFilters, Memory, MemoryUpdate, SearchHit


class VibememError(Exception):
    pass


class ConfigError(VibememError):
    pass


class ConnectionError(VibememError):
    pass


class NotFoundError(VibememError):
    pass


@dataclass(frozen=True)
class SearchRequest:
    query: str
    scope_type: str  # "global" | "project"
    scope_ids: Sequence[str]
    include_global: bool
    top_k: int = 8


class MemoryStore(Protocol):
    def add(self, memory: Memory) -> Memory: ...

    def get(self, memory_id: str) -> Memory: ...

    def edit(self, memory_id: str, update: MemoryUpdate) -> Memory: ...

    def delete(self, memory_id: str) -> None: ...

    def search(self, req: SearchRequest) -> list[SearchHit]: ...

    def list(self, filters: ListFilters) -> list[Memory]: ...


class MemoryCache(Protocol):
    def enabled(self) -> bool: ...

    def upsert(self, memory: Memory) -> None: ...

    def delete(self, memory_id: str) -> None: ...

    def search(self, req: SearchRequest) -> list[SearchHit]: ...

    def rebuild(self, memories: Sequence[Memory]) -> None: ...

    def count(self) -> Optional[int]: ...


from __future__ import annotations

from dataclasses import dataclass
from types import SimpleNamespace
from typing import Any

import pytest

from vibemem.store.base import SearchRequest
import vibemem.store.weaviate_store as ws


@dataclass(frozen=True)
class FakeFilter:
    op: str
    value: Any

    @classmethod
    def by_property(cls, name: str) -> "FakeFilter":
        return cls("prop", name)

    def equal(self, value: Any) -> "FakeFilter":
        assert self.op == "prop"
        return FakeFilter("eq", (self.value, value))

    @classmethod
    def all_of(cls, filters: list["FakeFilter"]) -> "FakeFilter":
        return cls("all", tuple(filters))

    @classmethod
    def any_of(cls, filters: list["FakeFilter"]) -> "FakeFilter":
        return cls("any", tuple(filters))


class StubQuery:
    def __init__(self) -> None:
        self.last_kwargs: dict[str, Any] = {}
        self.hybrid_calls = 0

    def hybrid(self, *args: Any, **kwargs: Any) -> Any:
        self.hybrid_calls += 1
        self.last_kwargs = dict(kwargs)
        return SimpleNamespace(objects=[])

    def bm25(self, *args: Any, **kwargs: Any) -> Any:
        self.last_kwargs = dict(kwargs)
        return SimpleNamespace(objects=[])


class StubQueryHybridNoVectorizer(StubQuery):
    def hybrid(self, *args: Any, **kwargs: Any) -> Any:
        self.hybrid_calls += 1
        raise RuntimeError("VectorFromInput was called without vectorizer on class VibeMemMemory")


def _make_store(query: StubQuery) -> ws.WeaviateStore:
    store = object.__new__(ws.WeaviateStore)
    store._collection = SimpleNamespace(query=query)
    return store


def test_weaviate_search_filter_includes_global(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(ws.wfilters, "Filter", FakeFilter)

    q = StubQuery()
    store = _make_store(q)

    req = SearchRequest(
        query="hello",
        scope_type="project",
        scope_ids=["repo::a", "repo"],
        include_global=True,
        top_k=5,
    )
    store.search(req)

    got = q.last_kwargs["filters"]
    expected = FakeFilter.any_of(
        [
            FakeFilter.all_of(
                [
                    FakeFilter.by_property("scope_type").equal("project"),
                    FakeFilter.any_of(
                        [
                            FakeFilter.by_property("scope_id").equal("repo::a"),
                            FakeFilter.by_property("scope_id").equal("repo"),
                        ]
                    ),
                ]
            ),
            FakeFilter.by_property("scope_type").equal("global"),
        ]
    )
    assert got == expected


def test_weaviate_search_filter_global_does_not_or_global(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(ws.wfilters, "Filter", FakeFilter)

    q = StubQuery()
    store = _make_store(q)

    req = SearchRequest(
        query="hello",
        scope_type="global",
        scope_ids=[],
        include_global=True,
        top_k=5,
    )
    store.search(req)

    got = q.last_kwargs["filters"]
    assert got == FakeFilter.by_property("scope_type").equal("global")


def test_weaviate_search_falls_back_to_bm25_when_no_vectorizer(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(ws.wfilters, "Filter", FakeFilter)

    q = StubQueryHybridNoVectorizer()
    store = _make_store(q)

    req = SearchRequest(
        query="hello",
        scope_type="project",
        scope_ids=["repo::a", "repo"],
        include_global=True,
        top_k=5,
    )
    store.search(req)

    assert q.hybrid_calls >= 1
    got = q.last_kwargs["filters"]
    expected = FakeFilter.any_of(
        [
            FakeFilter.all_of(
                [
                    FakeFilter.by_property("scope_type").equal("project"),
                    FakeFilter.any_of(
                        [
                            FakeFilter.by_property("scope_id").equal("repo::a"),
                            FakeFilter.by_property("scope_id").equal("repo"),
                        ]
                    ),
                ]
            ),
            FakeFilter.by_property("scope_type").equal("global"),
        ]
    )
    assert got == expected


def test_weaviate_search_uses_query_vector_when_embedder_present(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(ws.wfilters, "Filter", FakeFilter)

    class StubEmbedder:
        def embed(self, texts: list[str]) -> list[list[float]]:
            assert texts == ["hello"]
            return [[0.1, 0.2, 0.3]]

    q = StubQuery()
    store = _make_store(q)
    store._embedder = StubEmbedder()  # type: ignore[attr-defined]

    req = SearchRequest(query="hello", scope_type="global", scope_ids=[], include_global=False, top_k=5)
    store.search(req)

    assert "vector" in q.last_kwargs
    assert q.last_kwargs["vector"] == [0.1, 0.2, 0.3]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Optional, Sequence, Union

from vibemem.config import CacheMode, VibememConfig, load_config
from vibemem.models import Confidence, ListFilters, MemType, Memory, MemoryUpdate, SearchHit, ScopeType
from vibemem.scope import Granularity, ScopeInfo, derive_scope, scope_chain
from vibemem.store.base import ConnectionError, SearchRequest
from vibemem.store.chroma_cache import ChromaCache, chroma_present_on_disk, default_chroma_path
from vibemem.store.weaviate_store import WeaviateStore
from vibemem.util import normalize_where_paths, parse_csv_list, uniq


CacheModeOpt = Literal["auto", "on", "off"]
TagsInput = Union[str, Sequence[str]]


@dataclass(frozen=True)
class SearchResult:
    query: str
    scope: ScopeInfo
    scope_type: str
    scope_ids: list[str]
    include_global: bool
    include_parents: bool
    used: Literal["weaviate", "chroma"]
    hits: list[SearchHit]


def _cache_enabled(mode: CacheModeOpt, scope: ScopeInfo) -> tuple[bool, Path]:
    path = default_chroma_path(scope)
    if mode == "off":
        return False, path
    if mode == "on":
        return True, path
    return chroma_present_on_disk(path), path


def _open_cache(mode: CacheModeOpt, scope: ScopeInfo) -> ChromaCache:
    enabled, path = _cache_enabled(mode, scope)
    try:
        return ChromaCache(path=path, enabled=enabled)
    except ModuleNotFoundError as e:
        if enabled:
            raise ModuleNotFoundError(
                "Chroma cache requested but 'chromadb' is not installed. "
                "Install with: python -m pip install -U 'vibemem[cache]'"
            ) from e
        return ChromaCache(path=path, enabled=False)


def _apply_specificity_boost(hits: Sequence[SearchHit], scope_ids: list[str]) -> list[SearchHit]:
    if not scope_ids:
        return list(hits)

    idx: dict[str, int] = {sid: i for i, sid in enumerate(scope_ids)}
    denom = max(len(scope_ids) - 1, 1)

    out: list[SearchHit] = []
    for hit in hits:
        base = float(hit.score or 0.0)

        boost = 0.0
        mem = hit.memory
        if mem.scope_type.value == "project" and mem.scope_id in idx:
            specificity = (denom - idx[mem.scope_id]) / float(denom)
            boost = 0.05 * specificity
        elif mem.scope_type.value == "global":
            boost = 0.01

        out.append(hit.model_copy(update={"score": base + boost}))

    out.sort(key=lambda h: float(h.score or 0.0), reverse=True)
    return out


def _dedupe_hits_by_id(hits: Sequence[SearchHit]) -> list[SearchHit]:
    out: list[SearchHit] = []
    seen: set[str] = set()
    for hit in hits:
        if hit.id in seen:
            continue
        seen.add(hit.id)
        out.append(hit)
    return out


def _normalize_tags(tags: Optional[TagsInput]) -> list[str]:
    if tags is None:
        return []
    if isinstance(tags, str):
        return uniq(parse_csv_list(tags))
    return uniq([str(t).strip() for t in tags if str(t).strip()])


class VibeMem:
    def __init__(
        self,
        *,
        cfg: Optional[VibememConfig] = None,
        cwd: Optional[Path] = None,
        repo_root: Optional[Path] = None,
        scope_type: Optional[Literal["global", "project"]] = None,
        scope_id: Optional[str] = None,
        granularity: Granularity = "repo",
    ) -> None:
        self.cfg = cfg or load_config()
        self._default_granularity = granularity
        self.scope = derive_scope(
            cwd or Path.cwd(),
            repo_root_override=repo_root,
            scope_type_override=scope_type,
            scope_id_override=scope_id,
            granularity=granularity,
        )

    def scope_info(self) -> ScopeInfo:
        return self.scope

    def _resolve_scope(
        self,
        *,
        scope: Optional[Literal["global", "project"]] = None,
        scope_id: Optional[str] = None,
        cwd: Optional[Path] = None,
        repo_root: Optional[Path] = None,
        granularity: Optional[Granularity] = None,
    ) -> ScopeInfo:
        effective_cwd = cwd or self.scope.cwd
        effective_repo_root = repo_root if repo_root is not None else self.scope.repo_root
        effective_granularity = granularity or self._default_granularity
        return derive_scope(
            effective_cwd,
            repo_root_override=effective_repo_root,
            scope_type_override=scope,
            scope_id_override=scope_id,
            granularity=effective_granularity,
        )

    @staticmethod
    def _repo_slug_for_scope_id(scope_id: str) -> str:
        sid = (scope_id or "").strip()
        if not sid:
            return ""
        if "::" in sid:
            return sid.split("::", 1)[0].strip()
        return sid

    def search(
        self,
        query: str,
        *,
        top: int = 8,
        include_global: bool = True,
        include_parents: bool = True,
        cache: CacheModeOpt = "auto",
        scope: Optional[Literal["global", "project"]] = None,
        scope_id: Optional[str] = None,
        scope_ids: Optional[Sequence[str]] = None,
        cwd: Optional[Path] = None,
        repo_root: Optional[Path] = None,
        granularity: Optional[Granularity] = None,
    ) -> SearchResult:
        effective_scope = self._resolve_scope(
            scope=scope,
            scope_id=scope_id,
            cwd=cwd,
            repo_root=repo_root,
            granularity=granularity,
        )

        effective_scope_ids: list[str]
        if scope_ids is not None:
            effective_scope_ids = [str(s).strip() for s in scope_ids if str(s).strip()]
        elif effective_scope.scope_type == "project":
            effective_scope_ids = scope_chain(effective_scope, include_parents=include_parents)
        else:
            effective_scope_ids = []
            include_global = False

        if effective_scope.scope_type != "project":
            include_global = False

        cache_mode: CacheMode = cache if cache != "auto" else self.cfg.cache_mode
        chroma = _open_cache(cache_mode, effective_scope)

        req = SearchRequest(
            query=query,
            scope_type=effective_scope.scope_type,
            scope_ids=effective_scope_ids,
            include_global=include_global,
            top_k=top,
        )

        used: Literal["weaviate", "chroma"] = "weaviate"
        try:
            store = WeaviateStore(self.cfg)
            try:
                hits = store.search(req)
            finally:
                store.close()
        except ConnectionError:
            if chroma.enabled():
                used = "chroma"
                hits = chroma.search(req)
            else:
                raise

        boosted = _apply_specificity_boost(hits, effective_scope_ids)
        deduped = _dedupe_hits_by_id(boosted)[:top]

        return SearchResult(
            query=query,
            scope=effective_scope,
            scope_type=effective_scope.scope_type,
            scope_ids=effective_scope_ids,
            include_global=include_global,
            include_parents=include_parents,
            used=used,
            hits=deduped,
        )

    def search_hits(self, query: str, **kwargs: object) -> list[SearchHit]:
        return self.search(query, **kwargs).hits

    def search_memories(self, query: str, **kwargs: object) -> list[Memory]:
        return [h.memory for h in self.search(query, **kwargs).hits]

    def add_memory(
        self,
        *,
        mem_type: Union[MemType, str],
        text: str,
        tags: Optional[TagsInput] = None,
        confidence: Union[Confidence, str] = Confidence.med,
        verification: Optional[str] = None,
        errors: Optional[Sequence[str]] = None,
        files: Optional[Sequence[str]] = None,
        commands_run: Optional[Sequence[str]] = None,
        cache: CacheModeOpt = "auto",
        scope: Optional[Literal["global", "project"]] = None,
        scope_id: Optional[str] = None,
        cwd: Optional[Path] = None,
        repo_root: Optional[Path] = None,
        granularity: Optional[Granularity] = None,
    ) -> Memory:
        effective_scope = self._resolve_scope(
            scope=scope,
            scope_id=scope_id,
            cwd=cwd,
            repo_root=repo_root,
            granularity=granularity,
        )

        cache_mode: CacheMode = cache if cache != "auto" else self.cfg.cache_mode
        chroma = _open_cache(cache_mode, effective_scope)

        repo_slug = effective_scope.repo_slug
        rel_path = effective_scope.rel_path
        if effective_scope.scope_type == "project" and scope_id is not None:
            repo_slug = self._repo_slug_for_scope_id(scope_id) or repo_slug
            rel_path = ""
            if "::" in scope_id:
                rel_path = scope_id.split("::", 1)[1].strip().strip("/")

        m = Memory(
            text=text,
            mem_type=MemType(mem_type),
            tags=_normalize_tags(tags),
            scope_type=ScopeType(effective_scope.scope_type),
            scope_id=effective_scope.scope_id,
            repo=repo_slug,
            rel_path=rel_path,
            confidence=Confidence(confidence),
            verification=verification,
            error_signatures=list(errors) if errors else None,
            files=normalize_where_paths(files or []) or None,
            commands_run=list(commands_run) if commands_run else None,
        )

        store = WeaviateStore(self.cfg)
        try:
            created = store.add(m)
        finally:
            store.close()

        if chroma.enabled():
            chroma.upsert(created)

        return created

    def edit_memory(
        self,
        memory_id: str,
        *,
        text: Optional[str] = None,
        mem_type: Optional[Union[MemType, str]] = None,
        tags: Optional[TagsInput] = None,
        confidence: Optional[Union[Confidence, str]] = None,
        verification: Optional[str] = None,
        cache: CacheModeOpt = "auto",
        scope: Optional[Literal["global", "project"]] = None,
        scope_id: Optional[str] = None,
        cwd: Optional[Path] = None,
        repo_root: Optional[Path] = None,
        granularity: Optional[Granularity] = None,
    ) -> Memory:
        effective_scope = self._resolve_scope(
            scope=scope,
            scope_id=scope_id,
            cwd=cwd,
            repo_root=repo_root,
            granularity=granularity,
        )
        cache_mode: CacheMode = cache if cache != "auto" else self.cfg.cache_mode
        chroma = _open_cache(cache_mode, effective_scope)

        upd = MemoryUpdate(
            text=text,
            mem_type=MemType(mem_type) if mem_type is not None else None,
            tags=_normalize_tags(tags) if tags is not None else None,
            confidence=Confidence(confidence) if confidence is not None else None,
            verification=verification,
        )

        store = WeaviateStore(self.cfg)
        try:
            updated = store.edit(memory_id, upd)
        finally:
            store.close()

        if chroma.enabled():
            chroma.upsert(updated)

        return updated

    def delete_memory(
        self,
        memory_id: str,
        *,
        cache: CacheModeOpt = "auto",
        scope: Optional[Literal["global", "project"]] = None,
        scope_id: Optional[str] = None,
        cwd: Optional[Path] = None,
        repo_root: Optional[Path] = None,
        granularity: Optional[Granularity] = None,
    ) -> None:
        effective_scope = self._resolve_scope(
            scope=scope,
            scope_id=scope_id,
            cwd=cwd,
            repo_root=repo_root,
            granularity=granularity,
        )
        cache_mode: CacheMode = cache if cache != "auto" else self.cfg.cache_mode
        chroma = _open_cache(cache_mode, effective_scope)

        store = WeaviateStore(self.cfg)
        try:
            store.delete(memory_id)
        finally:
            store.close()

        if chroma.enabled():
            chroma.delete(memory_id)

    def get_memory(self, memory_id: str) -> Memory:
        store = WeaviateStore(self.cfg)
        try:
            return store.get(memory_id)
        finally:
            store.close()

    def list_memories(
        self,
        *,
        scope: Literal["global", "project", "all"] = "project",
        mem_type: Optional[Union[MemType, str]] = None,
        tag: Optional[str] = None,
        limit: int = 20,
        repo: Optional[str] = None,
        scope_ids: Optional[Sequence[str]] = None,
    ) -> list[Memory]:
        scope = scope.strip().lower()
        if scope not in ("global", "project", "all"):
            raise ValueError("scope must be 'global', 'project', or 'all'")

        effective_repo = repo
        if effective_repo is None and scope == "project" and self.scope.repo_slug:
            effective_repo = self.scope.repo_slug

        effective_scope_ids = list(scope_ids) if scope_ids is not None else None

        filters = ListFilters(
            scope=scope,
            mem_type=MemType(mem_type) if isinstance(mem_type, str) else mem_type,
            tag=(tag.strip() if tag else None),
            limit=limit,
            repo=effective_repo,
            scope_ids=effective_scope_ids,
        )

        store = WeaviateStore(self.cfg)
        try:
            return store.list(filters)
        finally:
            store.close()

    def sync_pull(self, *, limit: int = 200) -> dict[str, object]:
        if self.scope.scope_type != "project":
            raise ValueError("sync_pull requires project scope (run inside a repo or set scope_type='project').")

        chroma = _open_cache("on", self.scope)
        if not chroma.enabled():
            raise ModuleNotFoundError(
                "Chroma cache requested but 'chromadb' is not installed. "
                "Install with: python -m pip install -U 'vibemem[cache]'"
            )

        scope_ids = scope_chain(self.scope, include_parents=True)

        store = WeaviateStore(self.cfg)
        try:
            memories = store.list(
                ListFilters(
                    scope="project",
                    mem_type=None,
                    tag=None,
                    limit=limit,
                    repo=self.scope.repo_slug,
                    scope_ids=scope_ids,
                )
            )
        finally:
            store.close()

        chroma.rebuild(memories)
        return {
            "pulled": len(memories),
            "scope_ids": scope_ids,
            "chroma_path": str(default_chroma_path(self.scope)),
            "chroma_count": chroma.count(),
        }


def scope_info(
    *,
    cfg: Optional[VibememConfig] = None,
    cwd: Optional[Path] = None,
    repo_root: Optional[Path] = None,
    scope_type: Optional[Literal["global", "project"]] = None,
    scope_id: Optional[str] = None,
    granularity: Granularity = "repo",
) -> ScopeInfo:
    return VibeMem(
        cfg=cfg,
        cwd=cwd,
        repo_root=repo_root,
        scope_type=scope_type,
        scope_id=scope_id,
        granularity=granularity,
    ).scope_info()


def search(
    query: str,
    *,
    top: int = 8,
    include_global: bool = True,
    include_parents: bool = True,
    cache: CacheModeOpt = "auto",
    cfg: Optional[VibememConfig] = None,
    cwd: Optional[Path] = None,
    repo_root: Optional[Path] = None,
    scope_type: Optional[Literal["global", "project"]] = None,
    scope_id: Optional[str] = None,
    granularity: Granularity = "repo",
) -> SearchResult:
    return VibeMem(
        cfg=cfg,
        cwd=cwd,
        repo_root=repo_root,
        scope_type=scope_type,
        scope_id=scope_id,
        granularity=granularity,
    ).search(
        query,
        top=top,
        include_global=include_global,
        include_parents=include_parents,
        cache=cache,
    )


def search_hits(
    query: str,
    *,
    top: int = 8,
    include_global: bool = True,
    include_parents: bool = True,
    cache: CacheModeOpt = "auto",
    cfg: Optional[VibememConfig] = None,
    cwd: Optional[Path] = None,
    repo_root: Optional[Path] = None,
    scope_type: Optional[Literal["global", "project"]] = None,
    scope_id: Optional[str] = None,
    granularity: Granularity = "repo",
) -> list[SearchHit]:
    return search(
        query,
        top=top,
        include_global=include_global,
        include_parents=include_parents,
        cache=cache,
        cfg=cfg,
        cwd=cwd,
        repo_root=repo_root,
        scope_type=scope_type,
        scope_id=scope_id,
        granularity=granularity,
    ).hits


def search_memories(
    query: str,
    *,
    top: int = 8,
    include_global: bool = True,
    include_parents: bool = True,
    cache: CacheModeOpt = "auto",
    cfg: Optional[VibememConfig] = None,
    cwd: Optional[Path] = None,
    repo_root: Optional[Path] = None,
    scope_type: Optional[Literal["global", "project"]] = None,
    scope_id: Optional[str] = None,
    granularity: Granularity = "repo",
) -> list[Memory]:
    return [
        h.memory
        for h in search_hits(
        query,
        top=top,
        include_global=include_global,
        include_parents=include_parents,
        cache=cache,
        cfg=cfg,
        cwd=cwd,
        repo_root=repo_root,
        scope_type=scope_type,
        scope_id=scope_id,
        granularity=granularity,
    )
    ]


def add_memory(
    *,
    mem_type: Union[MemType, str],
    text: str,
    tags: Optional[TagsInput] = None,
    confidence: Union[Confidence, str] = Confidence.med,
    verification: Optional[str] = None,
    errors: Optional[Sequence[str]] = None,
    files: Optional[Sequence[str]] = None,
    commands_run: Optional[Sequence[str]] = None,
    cache: CacheModeOpt = "auto",
    cfg: Optional[VibememConfig] = None,
    cwd: Optional[Path] = None,
    repo_root: Optional[Path] = None,
    scope_type: Optional[Literal["global", "project"]] = None,
    scope_id: Optional[str] = None,
    granularity: Granularity = "repo",
) -> Memory:
    return VibeMem(
        cfg=cfg,
        cwd=cwd,
        repo_root=repo_root,
        scope_type=scope_type,
        scope_id=scope_id,
        granularity=granularity,
    ).add_memory(
        mem_type=mem_type,
        text=text,
        tags=tags,
        confidence=confidence,
        verification=verification,
        errors=errors,
        files=files,
        commands_run=commands_run,
        cache=cache,
    )


def edit_memory(
    memory_id: str,
    *,
    text: Optional[str] = None,
    mem_type: Optional[Union[MemType, str]] = None,
    tags: Optional[TagsInput] = None,
    confidence: Optional[Union[Confidence, str]] = None,
    verification: Optional[str] = None,
    cache: CacheModeOpt = "auto",
    cfg: Optional[VibememConfig] = None,
    cwd: Optional[Path] = None,
    repo_root: Optional[Path] = None,
    scope_type: Optional[Literal["global", "project"]] = None,
    scope_id: Optional[str] = None,
    granularity: Granularity = "repo",
) -> Memory:
    return VibeMem(
        cfg=cfg,
        cwd=cwd,
        repo_root=repo_root,
        scope_type=scope_type,
        scope_id=scope_id,
        granularity=granularity,
    ).edit_memory(
        memory_id,
        text=text,
        mem_type=mem_type,
        tags=tags,
        confidence=confidence,
        verification=verification,
        cache=cache,
    )


def delete_memory(
    memory_id: str,
    *,
    cache: CacheModeOpt = "auto",
    cfg: Optional[VibememConfig] = None,
    cwd: Optional[Path] = None,
    repo_root: Optional[Path] = None,
    scope_type: Optional[Literal["global", "project"]] = None,
    scope_id: Optional[str] = None,
    granularity: Granularity = "repo",
) -> None:
    VibeMem(
        cfg=cfg,
        cwd=cwd,
        repo_root=repo_root,
        scope_type=scope_type,
        scope_id=scope_id,
        granularity=granularity,
    ).delete_memory(memory_id, cache=cache)


def get_memory(
    memory_id: str,
    *,
    cfg: Optional[VibememConfig] = None,
    cwd: Optional[Path] = None,
    repo_root: Optional[Path] = None,
    scope_type: Optional[Literal["global", "project"]] = None,
    scope_id: Optional[str] = None,
    granularity: Granularity = "repo",
) -> Memory:
    return VibeMem(
        cfg=cfg,
        cwd=cwd,
        repo_root=repo_root,
        scope_type=scope_type,
        scope_id=scope_id,
        granularity=granularity,
    ).get_memory(memory_id)


def list_memories(
    *,
    scope: Literal["global", "project", "all"] = "project",
    mem_type: Optional[Union[MemType, str]] = None,
    tag: Optional[str] = None,
    limit: int = 20,
    repo: Optional[str] = None,
    scope_ids: Optional[Sequence[str]] = None,
    cfg: Optional[VibememConfig] = None,
    cwd: Optional[Path] = None,
    repo_root: Optional[Path] = None,
    scope_type: Optional[Literal["global", "project"]] = None,
    scope_id: Optional[str] = None,
    granularity: Granularity = "repo",
) -> list[Memory]:
    return VibeMem(
        cfg=cfg,
        cwd=cwd,
        repo_root=repo_root,
        scope_type=scope_type,
        scope_id=scope_id,
        granularity=granularity,
    ).list_memories(
        scope=scope,
        mem_type=mem_type,
        tag=tag,
        limit=limit,
        repo=repo,
        scope_ids=scope_ids,
    )


def sync_pull(
    *,
    limit: int = 200,
    cfg: Optional[VibememConfig] = None,
    cwd: Optional[Path] = None,
    repo_root: Optional[Path] = None,
    scope_type: Optional[Literal["global", "project"]] = None,
    scope_id: Optional[str] = None,
    granularity: Granularity = "repo",
) -> dict[str, object]:
    return VibeMem(
        cfg=cfg,
        cwd=cwd,
        repo_root=repo_root,
        scope_type=scope_type,
        scope_id=scope_id,
        granularity=granularity,
    ).sync_pull(limit=limit)



from __future__ import annotations

import pytest
from pydantic import ValidationError

from vibemem.models import Confidence, MemType, Memory, MemoryUpdate, ScopeType, memory_to_weaviate_properties


def test_memory_validation_requires_text() -> None:
    with pytest.raises(ValidationError):
        Memory(
            text="",
            mem_type=MemType.finding,
            tags=[],
            scope_type=ScopeType.project,
            scope_id="repo",
            repo="repo",
            rel_path="",
        )


def test_memory_dedupes_tags() -> None:
    m = Memory(
        text="hello",
        mem_type=MemType.finding,
        tags=["a", "a", " b ", ""],
        scope_type=ScopeType.project,
        scope_id="repo",
        repo="repo",
        rel_path="",
    )
    assert m.tags == ["a", "b"]


def test_memory_update_strips_text() -> None:
    u = MemoryUpdate(text="  hi  ")
    assert u.text == "hi"


def test_memory_to_weaviate_properties_omits_none_fields() -> None:
    m = Memory(
        text="hello",
        mem_type=MemType.finding,
        tags=[],
        scope_type=ScopeType.project,
        scope_id="repo",
        repo="repo",
        rel_path="",
        confidence=Confidence.med,
        verification=None,
        error_signatures=None,
        files=None,
        commands_run=None,
    )
    props = memory_to_weaviate_properties(m)
    assert "verification" not in props
    assert "error_signatures" not in props
    assert "files" not in props
    assert "commands_run" not in props